@@ -4,9 +4,16 @@ import re
 import aiohttp
 from typing import List
 import typing
+from .helpers import funcs
 import json
 from emoji import UNICODE_EMOJI
 
+def Validator(qtype: str):
+    try:
+        return funcs[qtype]
+    except KeyError:
+        raise InvalidFormType(f'Type {qtype} is invalid!')
+
 class FormResponse:
     def __init__(self,data:dict) -> None:
         for d in data.keys():
@@ -24,8 +31,10 @@ class Form:
         ctx (discord.ext.commands.Context): The context of the form.
 
         title (str): The title of the form.
+
+        cleanup (bool): Whether to cleanup and delete the form after finishing or not.
     """
-    def __init__(self, ctx:commands.Context, title):
+    def __init__(self, ctx:commands.Context, title, cleanup=False):
         self._ctx = ctx
         self._bot = ctx.bot
         self._questions = {}
@@ -34,9 +43,10 @@ class Form:
         self.editanddelete = False
         self.color = 0x2F3136
         self._incorrectmsg = None
-        self._retrymsg = None
+        self._retrymsg = "You have answered incorrectly."
         self._tries = None
         self.cancelkeywords = ['cancel', 'stop', 'quit']
+        self.cleanup = cleanup
 
     def set_timeout(self,timeout:int) -> None:
         """Sets the timeout for the form.
@@ -60,16 +70,30 @@ class Form:
         self._tries = tries
 
     def enable_cancelkeywords(self, enabled: bool):
+        """Enables or disables the cancellation of the form.
+
+        Parameters
+        ----------
+        enabled : bool
+            Whether the form is enabled or disabled.
+        """
         if enabled:
             self.cancelkeywords = ['cancel', 'stop', 'quit']
         else:
             self.cancelkeywords = []
 
-    def add_cancelkeyword(self, word):
+    def add_cancelkeyword(self, word: str):
+        """Adds a word that will cancel the form.
+
+        Parameters
+        ----------
+        word : str
+            The word to listen for.
+        """
         self.cancelkeywords.append(word.lower())
         return True
 
-    def add_question(self,question,key:str=None,qtype=None) -> List[dict]:
+    def add_question(self, question, key:str=None, qtype: List[typing.Union[Validator, typing.Callable]] = []) -> List[dict]:
         """Adds a question to the form.
         The valid qtypes are:
         `invite`,`channel`,`user`,`member`,`role`, and `category`
@@ -82,7 +106,7 @@ class Form:
         key : str
             What the attribute containing the answer should be called.
 
-        qtype : str, optional
+        qtype : List[Validator, function], optional
             The input validation to be used, by default None
 
         Returns
@@ -97,79 +121,23 @@ class Form:
             Is raised when the input validation type is invalid.
 
         """
+        if not isinstance(qtype, list):
+            qtype = [qtype]
         if not key:
             key = question
-        valid_qtypes = ['invite','channel','user','member','role','category','emoji']
         dictionary = {'res':None,'question':question}
-        if qtype:
-            dictionary['type'] = None
-            if qtype.lower() not in valid_qtypes:
-                raise InvalidFormType(f"Type '{qtype}' is invalid!")
-            dictionary['type'] = qtype
-            self._tries = 3
+        print(dictionary)
+        dictionary['type'] = qtype
+        print(dictionary)
+        self._tries = 3
 
         self._questions[key] = dictionary
+        print(self._questions)
         self.set_incorrect_message('You answered incorrectly too many times. Please try again.')
         self.set_retry_message(f'Please try again.')
+        print('-------------------------------------------')
         return self._questions
 
-    async def __validate_input(self,qtype,answer):
-        if qtype.lower() == 'invite':
-            try:
-                invite = await commands.InviteConverter().convert(self._ctx,answer)
-                return invite
-            except Exception as e:
-                return False
-        elif qtype.lower() == 'category':
-            try:
-                category = await commands.CategoryChannelConverter().convert(self._ctx, answer)
-                return category
-            except Exception:
-                return False
-        elif qtype.lower() == 'channel':
-            try:
-                channel = await commands.TextChannelConverter().convert(self._ctx,answer)
-                return channel
-            except:
-                return False
-        elif qtype.lower() == 'user':
-            try:
-                user = await commands.UserConverter().convert(self._ctx,answer)
-                return user
-            except:
-                return False
-        elif qtype.lower() == 'member':
-            try:
-                member = await commands.MemberConverter().convert(self._ctx,answer)
-                return member
-            except:
-                return False
-        elif qtype.lower() == 'role':
-            try:
-                role = await commands.RoleConverter().convert(self._ctx,answer)
-                return role
-            except:
-                return False
-        elif qtype.lower() == 'category':
-            try:
-                category = await commands.CategoryChannelConverter().convert(self._ctx,answer)
-                return category
-            except:
-                return False
-        elif qtype.lower() == 'emoji':
-            try:
-                emoji = await commands.EmojiConverter().convert(self._ctx,answer)
-                return emoji
-            except:
-                try:
-                    assert answer in UNICODE_EMOJI
-                except:
-                    return False
-                return answer
-        else:
-            self._tries -= 1
-            raise InvalidFormType(f"Type '{qtype}' is invalid!")
-
     def edit_and_delete(self,choice:bool=None) -> bool:
         """Toggles the edit and delete feature.
 
@@ -213,10 +181,9 @@ class Form:
         """
         self._incorrectmsg = message
 
-
     async def set_color(self,color:str) -> None:
         """Sets the color of the form embeds."""
-        match = re.match(r'(0x|#)(\d|(f|F|d|D|a|A|c|C)){6}', str(color))
+        match = re.match(r'(0x|#)(\d|(f|F|d|D|a|A|c|C|E|e|b|B)){6}', str(color))
         if not match:
             raise InvalidColor(f"{color} is invalid. Be sure to use colors such as '0xffffff'")
         if color.startswith("#"):
@@ -233,6 +200,9 @@ class Form:
         channel : discord.TextChannel, optional
             The channel to open the form in. If none, it is gotten from the context object set during initialization.
 
+        cleanup : bool
+            Whether to cleanup and delete the form after finishing or not.
+
         Returns
         -------
         FormResponse
@@ -275,44 +245,60 @@ class Form:
                 if self._questions[x]['question'].lower() == embed.description.lower():
                     question = x
                     nx = self._questions[x]
-
-            msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
-            ans = msg.content
-            if ans.lower() in self.cancelkeywords:
-                return None
-            if self.editanddelete:
-                await msg.delete()
-            key = question
-            if 'type' in self._questions[question].keys():
-                question = self._questions[question]
-                if 'type' in question.keys():
-                    while True:
-                        result = await self.__validate_input(question['type'],ans)
-                        if result:
-                            nx = question
+            while True:
+                msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
+                ans = msg.content
+                if ans.lower() in self.cancelkeywords:
+                    if self.cleanup:
+                        try:
+                            await msg.delete()
+                        except Exception:
+                            pass
+                        await prompt.delete()
+                    return None
+                if self.editanddelete:
+                    await msg.delete()
+                key = question
+                if 'type' in self._questions[question].keys():
+                    qinfo = self._questions[question]
+                    print(self._questions)
+                    for func in qinfo['type']:
+                        correct = False
+                        print(qinfo)
+                        print(type(func), func)
+                        result = await func(self._ctx, msg)
+                        print(bool(result is not None), result)
+                        if bool(result) is False:
+                            print('here')
+                            ot -= 1
+                            if ot <= 0:
+                                await channel.send(self._incorrectmsg)
+                                return None
+                            await channel.send(self._retrymsg + f" You have `{ot}` remaining.", delete_after=3)
+                        else:
+                            print('else')
+                            nx = qinfo
                             nx['res'] = result
                             self._questions[key] = nx
-                                    #self._questions[x] = ans
-                            #self._questions[question] = ans
-                            break
-                        else:
-                            self._tries -= 1
-                            await channel.send(self._retrymsg+f" You have `{self._tries}` remaining.",delete_after=3)
-                            msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
-                            ans = msg.content
-                            if self.editanddelete:
-                                await msg.delete()
+                            correct=True
                 else:
                     nx['res'] = ans
                     self._questions[key] = nx
-            else:
-                nx['res'] = ans
-                self._questions[key] = nx
+                    break
+                if correct:
+                    break
         for i in self._questions.keys():
             self._questions[i] = self._questions[i]
+        if self.cleanup:
+            try:
+                await msg.delete()
+            except Exception:
+                pass
+            await prompt.delete()
         return FormResponse(self._questions)
 
-class ContextAndChannelMissing(Exception):
+
+class MissingRequiredArgument(Exception):
     pass
 
 class NaiveForm:
@@ -327,11 +313,12 @@ class NaiveForm:
 
         bot (discord.ext.commands.Bot): The bot that will be running the form.
     """
-    def __init__(self,  title, channel: typing.Union[discord.abc.PrivateChannel, discord.abc.GuildChannel], bot: commands.Bot):
+    def __init__(self,  title, channel: typing.Union[discord.abc.PrivateChannel, discord.abc.GuildChannel], bot: commands.Bot, author: typing.Union[discord.Member, discord.User], cleanup=False):
         self._channel = channel
         self._bot = bot
         self._questions = {}
         self.title = title
+        self._author = author
         self.timeout = 120
         self.editanddelete = False
         self.color = 0x2F3136
@@ -339,14 +326,30 @@ class NaiveForm:
         self._retrymsg = None
         self._tries = None
         self.cancelkeywords = ['cancel', 'stop', 'quit']
+        self.cleanup=cleanup
 
     def enable_cancelkeywords(self, enabled: bool):
+        """Enables or disables the cancellation of the form.
+
+        Parameters
+        ----------
+        enabled : bool
+            Whether the form is enabled or disabled.
+        """
         if enabled:
             self.cancelkeywords = ['cancel', 'stop', 'quit']
+        else:
+            self.cancelkeywords = []
 
-    def add_cancelkeyword(self, word):
+    def add_cancelkeyword(self, word: str):
+        """Adds a word that will cancel the form.
+
+        Parameters
+        ----------
+        word : str
+            The word to listen for.
+        """
         self.cancelkeywords.append(word.lower())
-        return True
 
     def set_timeout(self,timeout:int) -> None:
         """Sets the timeout for the form.
@@ -399,7 +402,7 @@ class NaiveForm:
         """
         if not key:
             key = question
-        valid_qtypes = ['invite','channel', 'member', 'role', 'category', 'emoji']
+        valid_qtypes = ['invite','channel', 'member', 'role', 'category', 'emoji', 'file']
         dictionary = {'res':None,'question':question}
         if qtype:
             dictionary['type'] = None
@@ -413,7 +416,8 @@ class NaiveForm:
         self.set_retry_message(f'Please try again.')
         return self._questions
 
-    async def __validate_input(self,qtype,answer):
+    async def __validate_input(self,qtype,message):
+        answer = message.content
         if qtype.lower() == 'invite':
             async with aiohttp.ClientSession() as session:
                 r = await session.get(f'https://discord.com/api/invites/{answer}')
@@ -426,6 +430,11 @@ class NaiveForm:
             return bool(re.match(r'<@&\d{18}>', answer))
         elif qtype.lower() == 'emoji':
             return bool(re.match(r'<@&\d{18}>', answer)) or answer in UNICODE_EMOJI
+        elif qtype.lower() == "file":
+            try:
+                return message.attachments[0]
+            except IndexError:
+                return False
         else:
             raise InvalidFormType(f"Type '{qtype}' is invalid!")
 
@@ -479,7 +488,7 @@ class NaiveForm:
         if isinstance(color, discord.Color): self.color = color
         else: raise InvalidColor("This color is invalid! It should be a `discord.Color` instance.")
 
-    async def start(self, author: typing.Union[discord.Member, discord.User]) -> dict:
+    async def start(self,channel=None) -> dict:
         """Starts the form in the current channel.
 
         Parameters
@@ -487,6 +496,9 @@ class NaiveForm:
         channel : discord.TextChannel, optional
             The channel to open the form in. If none, it is gotten from the context object set during initialization.
 
+        cleanup : bool
+            Whether to cleanup and delete the form after finishing or not.
+
         Returns
         -------
         FormResponse
@@ -494,6 +506,9 @@ class NaiveForm:
         """
         elist = []
 
+        if not channel:
+            channel = self._channel
+
         qlist = []
         for n, q in enumerate(self._questions.values()):
             embed=discord.Embed(description=q['question'],color=self.color)
@@ -505,7 +520,6 @@ class NaiveForm:
 
             elist.append(embed)
 
-        channel = self._channel
 
         prompt = None
 
@@ -520,47 +534,63 @@ class NaiveForm:
             else:
                 prompt = await channel.send(embed=embed)
 
-            def check(m):
-                return m.channel == prompt.channel and m.author == author
+            def check(m: discord.Message):
+                return m.channel == prompt.channel and m.author == self._author
             question = None
             for x in self._questions.keys():
                 if self._questions[x]['question'].lower() == embed.description.lower():
                     question = x
                     nx = self._questions[x]
-
-            msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
-            ans = msg.content
-            if ans.lower() in self.cancelkeywords: return None
-            if self.editanddelete:
-                await msg.delete()
-            key = question
-            if 'type' in self._questions[question].keys():
-                question = self._questions[question]
-                if 'type' in question.keys():
-                    while True:
-                        result = await self.__validate_input(question['type'],ans)
-                        if result:
-                            nx = question
+            while True:
+                msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
+                ans = msg.content
+                if ans.lower() in self.cancelkeywords:
+                    if self.cleanup:
+                        try:
+                            await msg.delete()
+                        except Exception:
+                            pass
+                        await prompt.delete()
+                    return None
+                if self.editanddelete:
+                    await msg.delete()
+                key = question
+                if 'type' in self._questions[question].keys():
+                    qinfo = self._questions[question]
+                    print(self._questions)
+                    for func in qinfo['type']:
+                        correct = False
+                        print(qinfo)
+                        print(type(func), func)
+                        result = await func(self._channel, msg)
+                        print(bool(result is not None), result)
+                        if bool(result) is False:
+                            print('here')
+                            ot -= 1
+                            if ot <= 0:
+                                await channel.send(self._incorrectmsg)
+                                return None
+                            await channel.send(self._retrymsg + f" You have `{ot}` remaining.", delete_after=3)
+                        else:
+                            print('else')
+                            nx = qinfo
                             nx['res'] = result
                             self._questions[key] = nx
-                                    #self._questions[x] = ans
-                            #self._questions[question] = ans
-                            break
-                        else:
-                            self._tries -= 1
-                            await channel.send(self._retrymsg+f" You have `{self._tries}` remaining.",delete_after=3)
-                            msg = await self._bot.wait_for('message',check=check,timeout=self.timeout)
-                            ans = msg.content
-                            if self.editanddelete:
-                                await msg.delete()
+                            correct=True
                 else:
                     nx['res'] = ans
                     self._questions[key] = nx
-            else:
-                nx['res'] = ans
-                self._questions[key] = nx
+                    break
+                if correct:
+                    break
         for i in self._questions.keys():
             self._questions[i] = self._questions[i]
+        if self.cleanup:
+            try:
+                await msg.delete()
+            except Exception:
+                pass
+            await prompt.delete()
         return FormResponse(self._questions)
 
 
